@@ -72,11 +72,11 @@ if os.path.exists(LOGO_PATH):
 
 # -- Generate API ------------------------------------------------------------
 sys.path.insert(0, os.path.abspath("."))
-from _extend_docstrings import insert_math  # noqa: E402
+from _extend_docstrings import extend_docstrings  # noqa: E402
 from _relink_references import relink_references  # noqa: E402
 
+extend_docstrings()
 relink_references()
-insert_math()
 
 shutil.rmtree("api", ignore_errors=True)
 subprocess.call(
@@ -118,6 +118,8 @@ extensions = [
     "sphinx.ext.mathjax",
     "sphinx.ext.napoleon",
     "sphinx.ext.viewcode",
+    "sphinx_codeautolink",
+    "sphinx_comments",
     "sphinx_copybutton",
     "sphinx_panels",
     "sphinx_thebe",
@@ -137,10 +139,16 @@ add_module_names = False
 autodoc_default_options = {
     "exclude-members": ", ".join(
         [
+            "as_explicit",
             "default_assumptions",
             "doit",
             "evaluate",
             "is_commutative",
+            "is_extended_real",
+            "items",
+            "keys",
+            "precedence",
+            "values",
         ]
     ),
     "members": True,
@@ -149,17 +157,26 @@ autodoc_default_options = {
     "special-members": ", ".join(
         [
             "__call__",
-            "__getitem__",
         ]
     ),
 }
+autodoc_member_order = "bysource"
 autodoc_type_aliases = {
     "BuilderReturnType": "ampform.dynamics.builder.BuilderReturnType",
+    "FourMomenta": "ampform.kinematics.FourMomenta",
+    "FourMomentumSymbol": "ampform.kinematics.FourMomentumSymbol",
     "ParameterValue": "ampform.helicity.ParameterValue",
     "RangeDefinition": "symplot.RangeDefinition",
     "Slider": "symplot.Slider",
 }
 autodoc_typehints_format = "short"
+codeautolink_concat_default = True
+codeautolink_global_preface = """
+import numpy
+import numpy as np
+import sympy as sp
+from IPython.display import display
+"""
 AUTODOC_INSERT_SIGNATURE_LINEBREAKS = False
 graphviz_output_format = "svg"
 html_copy_source = True  # needed for download notebook button
@@ -191,6 +208,7 @@ html_theme_options = {
         "thebe": True,
         "thebelab": True,
     },
+    "show_navbar_depth": 2,
     "theme_dev_mode": True,
 }
 html_title = "AmpForm"
@@ -223,11 +241,13 @@ version_remapping = {
 def get_version(package_name: str) -> str:
     python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
     constraints_path = f"../.constraints/py{python_version}.txt"
+    package_name = package_name.lower()
     with open(constraints_path) as stream:
         constraints = stream.read()
     for line in constraints.split("\n"):
         line = line.split("#")[0]  # remove comments
         line = line.strip()
+        line = line.lower()
         if not line.startswith(package_name):
             continue
         if not line:
@@ -246,9 +266,27 @@ def get_version(package_name: str) -> str:
     return "stable"
 
 
+def get_minor_version(package_name: str) -> str:
+    installed_version = get_version(package_name)
+    if installed_version == "stable":
+        return installed_version
+    matches = re.match(r"^([0-9]+\.[0-9]+).*$", installed_version)
+    if matches is None:
+        raise ValueError(
+            "Could not find documentation for"
+            f" {package_name} v{installed_version}"
+        )
+    return matches[1]
+
+
 intersphinx_mapping = {
+    "IPython": (
+        f"https://ipython.readthedocs.io/en/{get_version('IPython')}",
+        None,
+    ),
     "attrs": (f"https://www.attrs.org/en/{get_version('attrs')}", None),
     "compwa-org": ("https://compwa-org.readthedocs.io", None),
+    "graphviz": ("https://graphviz.readthedocs.io/en/stable", None),
     "ipywidgets": (
         f"https://ipywidgets.readthedocs.io/en/{get_version('ipywidgets')}",
         None,
@@ -261,6 +299,7 @@ intersphinx_mapping = {
         f"https://mpl-interactions.readthedocs.io/en/{get_version('mpl-interactions')}",
         None,
     ),
+    "numpy": (f"https://numpy.org/doc/{get_minor_version('numpy')}", None),
     "pwa": ("https://pwa.readthedocs.io", None),
     "python": ("https://docs.python.org/3", None),
     "qrules": (
@@ -289,6 +328,7 @@ linkcheck_ignore = [
     "https://doi.org/10.1093/ptep/ptaa104",
     "https://home.fnal.gov/~kutschke/Angdist/angdist.ps",
     "https://hss-opus.ub.ruhr-uni-bochum.de",
+    "https://suchung.web.cern.ch",
 ]
 
 # Settings for myst_nb
@@ -335,6 +375,16 @@ interactively modify the parameters.
 }
 myst_update_mathjax = False
 
+# Settings for sphinx_comments
+comments_config = {
+    "hypothesis": True,
+    "utterances": {
+        "repo": f"ComPWA/{REPO_NAME}",
+        "issue-term": "pathname",
+        "label": "📝 Docs",
+    },
+}
+
 # Settings for Thebe cell output
 thebe_config = {
     "repository_url": html_theme_options["repository_url"],
