@@ -1,9 +1,9 @@
 # pylint: disable=no-member, no-self-use, redefined-outer-name
-# cspell:ignore atol doprint matexpr
-from textwrap import dedent
+# cspell:ignore atol doprint
+import inspect
+import textwrap
 from typing import Dict, Tuple
 
-import black
 import numpy as np
 import pytest
 import sympy as sp
@@ -12,29 +12,38 @@ from qrules.topology import Topology, create_isobar_topologies
 from sympy.printing.numpy import NumPyPrinter
 
 from ampform.kinematics import (
-    ArrayMultiplication,
-    ArraySum,
+    BoostZMatrix,
     Energy,
-    FourMomentumSymbols,
+    FourMomenta,
+    FourMomentumSymbol,
     FourMomentumX,
     FourMomentumY,
     FourMomentumZ,
     InvariantMass,
     Phi,
+    RotationYMatrix,
+    RotationZMatrix,
     Theta,
     ThreeMomentumNorm,
+    _ArraySize,
+    _OnesArray,
+    _ZerosArray,
     compute_helicity_angles,
     compute_invariant_masses,
     create_four_momentum_symbols,
     determine_attached_final_state,
 )
-from ampform.sympy._array_expressions import ArraySlice, ArraySymbol
+from ampform.sympy._array_expressions import (
+    ArrayMultiplication,
+    ArraySlice,
+    ArraySymbol,
+)
 
 
 @pytest.fixture(scope="session")
 def topology_and_momentum_symbols(
     data_sample: Dict[int, np.ndarray]
-) -> Tuple[Topology, FourMomentumSymbols]:
+) -> Tuple[Topology, FourMomenta]:
     n = len(data_sample)
     assert n == 4
     topologies = create_isobar_topologies(n)
@@ -45,53 +54,80 @@ def topology_and_momentum_symbols(
 
 @pytest.fixture(scope="session")
 def helicity_angles(
-    topology_and_momentum_symbols: Tuple[Topology, FourMomentumSymbols]
+    topology_and_momentum_symbols: Tuple[Topology, FourMomenta]
 ) -> Dict[str, sp.Expr]:
     topology, momentum_symbols = topology_and_momentum_symbols
     return compute_helicity_angles(momentum_symbols, topology)
 
 
-class TestArrayMultiplication:
-    def test_numpy_str(self):
-        n_events = 3
-        momentum = sp.MatrixSymbol("p", m=n_events, n=4)
-        beta = sp.Symbol("beta")
-        theta = sp.Symbol("theta")
-        expr = ArrayMultiplication(beta, theta, momentum)
-        numpy_code = _generate_numpy_code(expr)
-        numpy_code = black.format_str(
-            numpy_code, mode=black.Mode(line_length=70)
+class TestBoostZMatrix:
+    def test_boost_into_own_rest_frame_gives_mass(self):
+        p = FourMomentumSymbol("p")
+        n_events = _ArraySize(p)
+        beta = ThreeMomentumNorm(p) / Energy(p)
+        expr = BoostZMatrix(beta, n_events)
+        func = sp.lambdify(p, expr.doit())
+        p_array = np.array([[5, 0, 0, 1]])
+        boost_z = func(p_array)[0]
+        boosted_array = np.einsum("...ij,...j->...i", boost_z, p_array)
+        mass = 4.89897949
+        assert pytest.approx(boosted_array[0]) == [mass, 0, 0, 0]
+
+        expr = InvariantMass(p)
+        func = sp.lambdify(p, expr.doit())
+        mass_array = func(p_array)
+        assert pytest.approx(mass_array[0]) == mass
+
+    def test_numpycode_cse_in_expression_tree(self):
+        p, beta, phi, theta = sp.symbols("p beta phi theta")
+        expr = ArrayMultiplication(
+            BoostZMatrix(beta, n_events=_ArraySize(p)),
+            RotationYMatrix(theta, n_events=_ArraySize(p)),
+            RotationZMatrix(phi, n_events=_ArraySize(p)),
+            p,
         )
-        expected = """\
-        einsum("...ij,...j->...i", beta, einsum("...ij,...j->...i", theta, p))
+        func = sp.lambdify([], expr.doit(), cse=True)
+        src = inspect.getsource(func)
+        expected_src = """
+        def _lambdifygenerated():
+            x0 = 1/sqrt(1 - beta**2)
+            x1 = len(p)
+            x2 = ones(x1)
+            x3 = zeros(x1)
+            return (einsum("...ij,...jk,...kl,...l->...i", array(
+                    [
+                        [x0, x3, x3, -beta*x0],
+                        [x3, x2, x3, x3],
+                        [x3, x3, x2, x3],
+                        [-beta*x0, x3, x3, x0],
+                    ]
+                ).transpose((2, 0, 1)), array(
+                    [
+                        [x2, x3, x3, x3],
+                        [x3, cos(theta), x3, sin(theta)],
+                        [x3, x3, x2, x3],
+                        [x3, -sin(theta), x3, cos(theta)],
+                    ]
+                ).transpose((2, 0, 1)), array(
+                    [
+                        [x2, x3, x3, x3],
+                        [x3, cos(phi), -sin(phi), x3],
+                        [x3, sin(phi), cos(phi), x3],
+                        [x3, x3, x3, x2],
+                    ]
+                ).transpose((2, 0, 1)), p))
         """
-        assert numpy_code == dedent(expected)
-
-
-class TestArraySum:
-    def test_latex(self):
-        x, y = sp.symbols("x y")
-        array_sum = ArraySum(x ** 2, sp.cos(y))
-        assert sp.latex(array_sum) == R"x^{2} + \cos{\left(y \right)}"
-
-    def test_latex_array_symbols(self):
-        p0, p1, p2, p3 = sp.symbols("p:4", cls=ArraySymbol)
-        array_sum = ArraySum(p0, p1, p2, p3)
-        assert sp.latex(array_sum) == "{p}_{0123}"
-
-    def test_numpy(self):
-        expr = ArraySum(*sp.symbols("x y"))
-        numpy_code = _generate_numpy_code(expr)
-        assert numpy_code == "x + y"
+        expected_src = textwrap.dedent(expected_src)
+        assert src.strip() == expected_src.strip()
 
 
 class TestFourMomentumXYZ:
     def symbols(
         self,
     ) -> Tuple[
-        ArraySymbol, Energy, FourMomentumX, FourMomentumY, FourMomentumZ
+        FourMomentumSymbol, Energy, FourMomentumX, FourMomentumY, FourMomentumZ
     ]:
-        p = ArraySymbol("p")
+        p = FourMomentumSymbol("p")
         e = Energy(p)
         p_x = FourMomentumX(p)
         p_y = FourMomentumY(p)
@@ -132,7 +168,7 @@ class TestInvariantMass:
         state_id: int,
         expected_mass: float,
     ):
-        p = ArraySymbol(f"p{state_id}")
+        p = FourMomentumSymbol(f"p{state_id}")
         mass = InvariantMass(p)
         np_mass = sp.lambdify(p, mass.doit(), "numpy")
         four_momenta = data_sample[state_id]
@@ -144,7 +180,7 @@ class TestInvariantMass:
 class TestThreeMomentumNorm:
     @property
     def p_norm(self) -> ThreeMomentumNorm:
-        p = ArraySymbol("p")
+        p = FourMomentumSymbol("p")
         return ThreeMomentumNorm(p)
 
     def test_latex(self):
@@ -159,7 +195,7 @@ class TestThreeMomentumNorm:
 class TestPhi:
     @property
     def phi(self) -> Theta:
-        p = ArraySymbol("p")
+        p = FourMomentumSymbol("p")
         return Phi(p)
 
     def test_latex(self):
@@ -175,7 +211,7 @@ class TestPhi:
 class TestTheta:
     @property
     def theta(self) -> Theta:
-        p = ArraySymbol("p")
+        p = FourMomentumSymbol("p")
         return Theta(p)
 
     def test_latex(self):
@@ -191,6 +227,129 @@ class TestTheta:
         )
 
 
+class TestRotationYMatrix:
+    @pytest.fixture(scope="session")
+    def rotation_expr(self):
+        angle, n_events = sp.symbols("a n")
+        return RotationYMatrix(angle, n_events)
+
+    @pytest.fixture(scope="session")
+    def rotation_func(self, rotation_expr):
+        angle = sp.Symbol("a")
+        rotation_expr = rotation_expr.doit()
+        rotation_expr = rotation_expr.subs(sp.Symbol("n"), _ArraySize(angle))
+        return sp.lambdify(angle, rotation_expr, cse=True)
+
+    def test_numpycode_cse(self, rotation_expr: RotationYMatrix):
+        func = sp.lambdify([], rotation_expr.doit(), cse=True)
+        src = inspect.getsource(func)
+        expected_src = """
+        def _lambdifygenerated():
+            return (array(
+                    [
+                        [ones(n), zeros(n), zeros(n), zeros(n)],
+                        [zeros(n), cos(a), zeros(n), sin(a)],
+                        [zeros(n), zeros(n), ones(n), zeros(n)],
+                        [zeros(n), -sin(a), zeros(n), cos(a)],
+                    ]
+                ).transpose((2, 0, 1)))
+        """
+        expected_src = textwrap.dedent(expected_src)
+        assert src.strip() == expected_src.strip()
+
+    def test_rotation_over_pi_flips_xz(self, rotation_func):
+        vectors = np.array([[1, 1, 1, 1]])
+        angle_array = np.array([np.pi])
+        rotated_vectors = np.einsum(
+            "...ij,...j->...j", rotation_func(angle_array), vectors
+        )
+        assert pytest.approx(rotated_vectors) == np.array([[1, -1, 1, -1]])
+
+
+class TestRotationZMatrix:
+    @pytest.fixture(scope="session")
+    def rotation_expr(self):
+        angle, n_events = sp.symbols("a n")
+        return RotationZMatrix(angle, n_events)
+
+    @pytest.fixture(scope="session")
+    def rotation_func(self, rotation_expr):
+        angle = sp.Symbol("a")
+        rotation_expr = rotation_expr.doit()
+        rotation_expr = rotation_expr.subs(sp.Symbol("n"), _ArraySize(angle))
+        return sp.lambdify(angle, rotation_expr, cse=True)
+
+    def test_numpycode_cse(self, rotation_expr: RotationZMatrix):
+        func = sp.lambdify([], rotation_expr.doit(), cse=True)
+        src = inspect.getsource(func)
+        expected_src = """
+        def _lambdifygenerated():
+            return (array(
+                    [
+                        [ones(n), zeros(n), zeros(n), zeros(n)],
+                        [zeros(n), cos(a), -sin(a), zeros(n)],
+                        [zeros(n), sin(a), cos(a), zeros(n)],
+                        [zeros(n), zeros(n), zeros(n), ones(n)],
+                    ]
+                ).transpose((2, 0, 1)))
+        """
+        expected_src = textwrap.dedent(expected_src)
+        assert src.strip() == expected_src.strip()
+
+    def test_rotation_over_pi_flips_xy(self, rotation_func):
+        vectors = np.array([[1, 1, 1, 1]])
+        angle_array = np.array([np.pi])
+        rotated_vectors = np.einsum(
+            "...ij,...j->...j", rotation_func(angle_array), vectors
+        )
+        assert pytest.approx(rotated_vectors) == np.array([[1, -1, -1, 1]])
+
+
+@pytest.mark.parametrize("rotation", [RotationYMatrix, RotationZMatrix])
+def test_rotation_latex_repr_is_identical_with_doit(rotation):
+    angle, n_events = sp.symbols("a n")
+    expr = rotation(angle, n_events)
+    assert sp.latex(expr) == sp.latex(expr.doit())
+
+
+@pytest.mark.parametrize("rotation", [RotationYMatrix, RotationZMatrix])
+def test_rotation_over_multiple_two_pi_is_identity(rotation):
+    angle = sp.Symbol("a")
+    expr = rotation(angle)
+    func = sp.lambdify(angle, expr.doit(), cse=True)
+    angle_array = np.arange(-2, 4, 1) * 2 * np.pi
+    rotation_matrices = func(angle_array)
+    identity = np.array(
+        [
+            [1, 0, 0, 0],
+            [0, 1, 0, 0],
+            [0, 0, 1, 0],
+            [0, 0, 0, 1],
+        ]
+    )
+    identity = np.tile(identity, reps=(len(angle_array), 1, 1))
+    assert pytest.approx(rotation_matrices) == identity
+
+
+class TestOnesZerosArray:
+    @pytest.mark.parametrize("array_type", ["ones", "zeros"])
+    @pytest.mark.parametrize("shape", [10, (4, 2), [3, 5, 7]])
+    def test_numpycode(self, array_type, shape):
+        if array_type == "ones":
+            expr_class = _OnesArray
+            array_func = np.ones
+        elif array_type == "zeros":
+            expr_class = _ZerosArray
+            array_func = np.zeros
+        else:
+            raise NotImplementedError
+        array_expr = expr_class(shape)
+        create_array = sp.lambdify([], array_expr)
+        array = create_array()
+        np.testing.assert_array_equal(array, array_func(shape))
+
+
+@pytest.mark.parametrize("use_cse", [False, True])
 @pytest.mark.parametrize(
     ("angle_name", "expected_values"),
     [
@@ -298,9 +457,10 @@ class TestTheta:
         ),
     ],
 )
-def test_compute_helicity_angles(
+def test_compute_helicity_angles(  # pylint: disable=too-many-arguments
+    use_cse: bool,
     data_sample: Dict[int, np.ndarray],
-    topology_and_momentum_symbols: Tuple[Topology, FourMomentumSymbols],
+    topology_and_momentum_symbols: Tuple[Topology, FourMomenta],
     angle_name: str,
     expected_values: np.ndarray,
     helicity_angles: Dict[str, sp.Expr],
@@ -308,13 +468,13 @@ def test_compute_helicity_angles(
     _, momentum_symbols = topology_and_momentum_symbols
     four_momenta = data_sample.values()
     expr = helicity_angles[angle_name]
-    np_angle = sp.lambdify(momentum_symbols.values(), expr.doit())
+    np_angle = sp.lambdify(momentum_symbols.values(), expr.doit(), cse=use_cse)
     computed = np_angle(*four_momenta)
     np.testing.assert_allclose(computed, expected_values, atol=1e-5)
 
 
 def test_compute_invariant_masses_names(
-    topology_and_momentum_symbols: Tuple[Topology, FourMomentumSymbols]
+    topology_and_momentum_symbols: Tuple[Topology, FourMomenta]
 ):
     topology, momentum_symbols = topology_and_momentum_symbols
     invariant_masses = compute_invariant_masses(momentum_symbols, topology)
@@ -331,7 +491,7 @@ def test_compute_invariant_masses_names(
 
 def test_compute_invariant_masses_single_mass(
     data_sample: Dict[int, np.ndarray],
-    topology_and_momentum_symbols: Tuple[Topology, FourMomentumSymbols],
+    topology_and_momentum_symbols: Tuple[Topology, FourMomenta],
 ):
     topology, momentum_symbols = topology_and_momentum_symbols
     momentum_values = data_sample.values()
@@ -348,7 +508,7 @@ def test_compute_invariant_masses_single_mass(
 def test_compute_invariant_masses(
     mass_name: str,
     data_sample: Dict[int, np.ndarray],
-    topology_and_momentum_symbols: Tuple[Topology, FourMomentumSymbols],
+    topology_and_momentum_symbols: Tuple[Topology, FourMomenta],
 ):
     topology, momentum_symbols = topology_and_momentum_symbols
     momentum_values = data_sample.values()
@@ -366,7 +526,7 @@ def test_compute_invariant_masses(
 def __compute_mass(array: np.ndarray) -> np.ndarray:
     energy = array[:, 0]
     three_momentum = array[:, 1:]
-    mass_squared = energy ** 2 - np.sum(three_momentum ** 2, axis=1)
+    mass_squared = energy**2 - np.sum(three_momentum**2, axis=1)
     return complex_sqrt(mass_squared)
 
 
